/*++

Module Name:

    ioctl.c


X36 IO handler.

Overview:

	This module is responsible for handling all the device control calls that
pass through this filter.  It performs the following functions:
	-> Tweaking the USB descriptors as they come up through the stack to add
	   the parameters necessary to create a virtual keyboard.
	-> Capturing an I/O request to reply with keyboard data on my own timers
	-> Replying with keyboard data
	-> Filter X36 joystick data

--*/



#include <WDM.H>
#include "usbioctl.h"
#include "usbdi.h"
#include "usbdlib.h"
#include "filter.h"
#include "fakehid.h"


// This works but these should really be in the device extension.




#ifdef ALLOC_PRAGMA
        #pragma alloc_text(PAGE, VA_Ioctl)
#endif


NTSTATUS VA_Ioctl(struct DEVICE_EXTENSION *  devExt, PIRP Irp)
{
    NTSTATUS            ntStatus = STATUS_SUCCESS;    
    PIO_STACK_LOCATION  IrpStack = IoGetCurrentIrpStackLocation(Irp);    
    ULONG dwControlCode = IrpStack->Parameters.DeviceIoControl.IoControlCode;
	PURB pUrb = (PURB)IrpStack->Parameters.Others.Argument1;
	KIRQL OldIrqL;
			    
    if(IOCTL_INTERNAL_USB_SUBMIT_URB == dwControlCode)
	{

		// DBGOUT(("VA_Ioctl: %d", IrpStack->Parameters.DeviceIoControl.IoControlCode));
		if (pUrb->UrbHeader.Function == URB_FUNCTION_BULK_OR_INTERRUPT_TRANSFER)
		{
			
			if ((devExt->ReadIrp == NULL)&&(KeGetCurrentIrql()==DISPATCH_LEVEL)) // Don't capture if we're not really running.
			{
				KdPrint(("Capturing keyboard read IRP.\n"));
				
				KeAcquireSpinLockAtDpcLevel(&devExt->ReadLock);
				devExt->ReadIrp = Irp;
				KeReleaseSpinLockFromDpcLevel(&devExt->ReadLock);
				
	            IncrementPendingActionCount(devExt);
	            
	            IoMarkIrpPending(Irp);
	          	ntStatus = STATUS_PENDING;
	          	
	          	if (devExt->NewData)
	          	{
//	          		DbgPrint("Filling request right away.  New data: %d\n",devExt->NewData);
	          		KeInsertQueueDpc(&devExt->ReturnDPC,NULL,NULL);
	          	}
            }
            else
            {
            	IoCopyCurrentIrpStackLocationToNext(Irp);
	    		IoSetCompletionRoutine(Irp, &IoCtlCompletion, NULL, TRUE, TRUE, TRUE);    
	    		ntStatus = IoCallDriver(devExt->topDevObj, Irp);
	    	}
		}
		else 
		{
			// Capture downgoing time for lag control.
			// KeQuerySystemTime(&devExt->StartTime);
			
			
			KeAcquireSpinLockAtDpcLevel(&devExt->ReadLock);
			if ( devExt->ReadIrp != NULL )
			{		
				if (pUrb->UrbHeader.Function == URB_FUNCTION_RESET_PIPE)
				{
					KeCancelTimer(&devExt->Timer);
					KeReleaseSpinLockFromDpcLevel(&devExt->ReadLock);
					CompleteRead( NULL, devExt, NULL, NULL);				
				}
				else if (pUrb->UrbHeader.Function == URB_FUNCTION_ABORT_PIPE)
				{
					KeCancelTimer(&devExt->Timer);				
				   	devExt->ReadIrp->IoStatus.Status = STATUS_CANCELLED;
				    IoCompleteRequest(devExt->ReadIrp, IO_NO_INCREMENT);	
					devExt->ReadIrp = NULL;
					KeReleaseSpinLockFromDpcLevel(&devExt->ReadLock);
    				DecrementPendingActionCount( devExt );										
				}
				else
					KeReleaseSpinLockFromDpcLevel(&devExt->ReadLock);
			}
			else
				KeReleaseSpinLockFromDpcLevel(&devExt->ReadLock);

				
	    	// Pass the request down and pick it up on the way back.
	    	IoCopyCurrentIrpStackLocationToNext(Irp);
	    	IoSetCompletionRoutine(Irp, &IoCtlCompletion, NULL, TRUE, TRUE, TRUE);    
	    	ntStatus = IoCallDriver(devExt->topDevObj, Irp);
	    }
	}
	else
	{
		IoSkipCurrentIrpStackLocation(Irp);
		ntStatus = IoCallDriver(devExt->topDevObj, Irp);
	}

    return ntStatus;    
}



NTSTATUS IoCtlCompletion( IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp, IN PVOID Context )
{
	/* Should print the data in the system. */
    NTSTATUS            ntStatus = STATUS_SUCCESS;
    struct DEVICE_EXTENSION *  DeviceExtension = DeviceObject->DeviceExtension;
    PIO_STACK_LOCATION  IrpStack = IoGetCurrentIrpStackLocation(Irp);
    
   	PURB pUrb = (PURB)IrpStack->Parameters.Others.Argument1;
   	struct _URB_CONTROL_TRANSFER   *pControlTransfer;
	struct _URB_BULK_OR_INTERRUPT_TRANSFER *pBulkOrInterruptTransfer;
   	BOOLEAN bReadFromDevice;
   	PUCHAR BufPtr=NULL;
   	PUCHAR Temp;
   	LARGE_INTEGER LocalTime;
   	PQENTRY Current;
   	unsigned char Delta[0x0b];
   	unsigned char switch1,switch2,shift;
    unsigned long b,i,pos;
    
    
    if (Irp->PendingReturned) {
        IoMarkIrpPending( Irp );
    }
    
    
    // Lag avoidance routine.
	KeQuerySystemTime(&LocalTime);
	

	// OK,we need to (a) snapture the report id; then (b) snapture the interrupts themselves.
	// To get the report ID we need to know which one to watch for.
	// So, first we have to capture the HID descriptor and parse that.  It comes between the configuration and interface descriptors.


	if (pUrb != NULL)
	{
		USHORT wFunction = pUrb->UrbHeader.Function;
		USHORT wLength = pUrb->UrbHeader.Length;
		USBD_STATUS lUsbdStatus = pUrb->UrbHeader.Status;
		

	
		

		switch(wFunction)
		{			
			case URB_FUNCTION_CONTROL_TRANSFER:
				 pControlTransfer = (struct _URB_CONTROL_TRANSFER *) pUrb;
				 bReadFromDevice = (BOOLEAN)(pControlTransfer->TransferFlags & USBD_TRANSFER_DIRECTION_IN);
				
				if	(bReadFromDevice)
				{
					// Figure out what packet this was.
					KdPrint(("  SetupPacket          :"));
					for(b=0; b<sizeof(pControlTransfer->SetupPacket); b++)
						KdPrint((" %02x", pControlTransfer->SetupPacket[b]));
					KdPrint(("\n"));

					if ((pControlTransfer->SetupPacket[0]&0x80)&&(pControlTransfer->SetupPacket[1]==0x06))
					{
						
						BufPtr=NULL;
						if ((PUCHAR)pControlTransfer->TransferBuffer)
						{
							BufPtr=(PUCHAR)pControlTransfer->TransferBuffer;
						}
						else if(pControlTransfer->TransferBufferMDL)
						{
							BufPtr = (PUCHAR)MmGetSystemAddressForMdl(pControlTransfer->TransferBufferMDL);
						}
						if (BufPtr==NULL)
						{	// Confused.  Abort.
							KdPrint(("Can't find buffer, aborting.\n"));
							return STATUS_SUCCESS;
						}
												
						if ((pControlTransfer->SetupPacket[3]==0x02)&&((pControlTransfer->SetupPacket[6]>0x09)||(pControlTransfer->SetupPacket[7]>0)))
						{
							KdPrint(("Complete descriptor transferred.  Locating HID descriptor.\n"));
							
							for(pos=0;(BufPtr[pos+1]!=0x21)&&(pos<pControlTransfer->TransferBufferLength);pos=pos+BufPtr[pos]);
							
							if (pos>=pControlTransfer->TransferBufferLength)
							{
								KdPrint(("Can't find HID descriptor, aborting.\n"));
								return STATUS_SUCCESS;
							}
							
							DeviceExtension->ReportDescType = BufPtr[pos+6];
							DeviceExtension->ReportBufSize = BufPtr[pos+7];
							BufPtr[pos+7]+=sizeof(MyReportDescriptor)+2;
														
														
								
							if (Temp = (PUCHAR)MmGetSystemAddressForMdl(pControlTransfer->TransferBufferMDL))
								RtlCopyMemory(Temp,BufPtr,pControlTransfer->TransferBufferLength);		
						}
						else if (pControlTransfer->SetupPacket[3]==DeviceExtension->ReportDescType)
						{
							KdPrint(("Report Descriptor.\n"));
							
							// Fix axis maxima/minima.
							pos=9;
							for(i=0; i<6; i++)
							{
								if (DeviceExtension->Configurations[0][0][0]->Axis[i].Min)
									BufPtr[pos]=(unsigned char)DeviceExtension->Configurations[0][0][0]->Axis[i].Min;
								pos+=(i<2)?3:2;
								if (DeviceExtension->Configurations[0][0][0]->Axis[i].Max)
									*(unsigned short*)(&BufPtr[pos])=(unsigned short)DeviceExtension->Configurations[0][0][0]->Axis[i].Max;
								pos+=10;
							}
							
							// Fix throttle type.
							BufPtr[35]=0x32;
							
							
							RtlMoveMemory(BufPtr+8,BufPtr+6,pControlTransfer->TransferBufferLength-6);
							
							BufPtr[6]=0x85;			BufPtr[7]=0x01;	
							pControlTransfer->TransferBufferLength+=2;
							RtlCopyMemory(&BufPtr[pControlTransfer->TransferBufferLength],MyReportDescriptor, sizeof(MyReportDescriptor));
							pControlTransfer->TransferBufferLength+=sizeof(MyReportDescriptor);
							
						}
					}
				}
				break;


			case URB_FUNCTION_BULK_OR_INTERRUPT_TRANSFER:
				// This is the stick reporting its packet.
				// KdPrint(("Stick reporting packet.\n"));
				
				pBulkOrInterruptTransfer = (struct _URB_BULK_OR_INTERRUPT_TRANSFER *) pUrb;
								
				
				// Do nothing at this point.  Later we'll mask off buttons.  Need to parse report descriptor first.
				// Force mask off buttons immediately after the ones lit in the packets.
				if (pBulkOrInterruptTransfer->TransferBufferLength == 0x0b)
				{
					BufPtr = NULL;
				
					if ((PUCHAR)pBulkOrInterruptTransfer->TransferBuffer)
					{
						BufPtr=(PUCHAR)pBulkOrInterruptTransfer->TransferBuffer;
					}
					else if(pBulkOrInterruptTransfer->TransferBufferMDL)
					{
						BufPtr = (PUCHAR)MmGetSystemAddressForMdl(pBulkOrInterruptTransfer->TransferBufferMDL);
					}
					if (BufPtr==NULL)
					{	// Confused.  Abort.
						return STATUS_SUCCESS;
					}						
					
					
					// Compute data structures.
					// Store system time.
					KeQuerySystemTime(&LocalTime);
												
					
					// Expand hat1.
					switch (BufPtr[0x0a]&0xF0)
					{
						case 0:
							// Do nothing. ;)
							break;
						case 0x10:
							BufPtr[0x0a]=(BufPtr[0x0b]&0x0F)+0x04;
							break;
						case 0x20:
							BufPtr[0x0a]=(BufPtr[0x0b]&0x0F)+0x04+0x08;
							break;
						case 0x30:
							BufPtr[0x0a]=(BufPtr[0x0b]&0x0F)+0x08;
							break;
						case 0x40:
							BufPtr[0x0a]=(BufPtr[0x0b]&0x0F)+0x08+0x10;
							break;
						case 0x50:
							BufPtr[0x0a]=(BufPtr[0x0b]&0x0F)+0x10;
							break;
						case 0x60:
							BufPtr[0x0a]=(BufPtr[0x0b]&0x0F)+0x10+0x20;
							break;
						case 0x70:
							BufPtr[0x0a]=(BufPtr[0x0b]&0x0F)+0x20;
							break;
						case 0x80:
							BufPtr[0x0a]=(BufPtr[0x0b]&0x0F)+0x20+0x04;
							break;
					}		
					
					
					// Compute deltas.
					for(i=0;i<0x0b;i++)
					{
						Delta[i]=DeviceExtension->LastPosition[i] ^ BufPtr[i];
						DeviceExtension->LastPosition[i] = BufPtr[i];
					}
					
					// Load configurations.
					if (BufPtr[8]&0x01)
						switch1 = 1;
					else if (BufPtr[8]&0x02)
						switch1 = 2;
					else
						switch1 = 3;
						
					if (BufPtr[8]&0x08)
						switch2 = 1;
					else if (BufPtr[8]&0x10)
						switch2 = 2;
					else
						switch2 = 3;
						
					shift = (BufPtr[7]&0x40)?2:1;
						

					// Data structures computed.  Begin engine.

					pos=*(unsigned long*)&BufPtr[7];
//					DbgPrint("Buttons: %X ",pos);
					pos=*(unsigned long*)&Delta[7];
//					DbgPrint("Delta: %X   Mode: %d   Aux: %d    Shift: %d\n",pos,switch1,switch2,shift);
					
					// Script queueing engine.
					// Buttons.
					for(i=0;i<8;i++)
					{
						if ((BufPtr[7]>>i)&0x01) // Button is active
						{
							if ((Delta[7]>>i)&0x01) // Button toggled
							{
								// Button pressed.
								AddAllScripts(DeviceExtension,switch1,switch2,shift,&(DeviceExtension->Configurations[0][0][0]->Button[i].Pressed),LocalTime);
							}
							else
							{
								// Button held.
							}
						}
						else if ((Delta[7]>>i)&0x01) // Button toggled
						{
							// Button released.
							AddAllScripts(DeviceExtension,switch1,switch2,shift,&(DeviceExtension->Configurations[0][0][0]->Button[i].Released),LocalTime);
						}
					}

					// Switches.
					// Each 3-way pressed, released, and hold only has one value per configuration.
					// We need to identify the previous configuration with regards to the switch in question so we can perform its release.
					for(i=0;i<2;i++)
					{
						if ((Delta[8]>>(i*3))&0x07) // State has changed.  Will be two deltas.  Deltas ^ current = released, delta & current = pressed.
						{
							unsigned char pressed, released;
							pressed = ((Delta[8]>>(i*3))&0x07) & ((BufPtr[8]>>(i*3))&0x07);
							pressed = ((pressed > 3)||(pressed<1)) ? 3 : pressed;
							released = ((Delta[8]>>(i*3))&0x07) ^ ((BufPtr[8]>>(i*3))&0x07);
							released = ((released > 3)||(released<1)) ? 3 : released;
//							DbgPrint("Switch %d:   Pressed: %d   Released: %d\n",i,pressed,released);
							if (i==0) // Stupid but what can I do?  0 is the mode switch == switch1.
							{
								AddAllScripts(DeviceExtension,pressed,switch2,shift,&DeviceExtension->Configurations[0][0][0]->Slider[i].Pressed,LocalTime);
								AddAllScripts(DeviceExtension,released,switch2,shift,&DeviceExtension->Configurations[0][0][0]->Slider[i].Released,LocalTime);
							}
							else
							{
								AddAllScripts(DeviceExtension,switch1,pressed,shift,&DeviceExtension->Configurations[0][0][0]->Slider[i].Pressed,LocalTime);
								AddAllScripts(DeviceExtension,switch1,released,shift,&DeviceExtension->Configurations[0][0][0]->Slider[i].Released,LocalTime);
							}
						}
						else
						{
							// Slider held.
							;
						}
					}



					for(i=0;i<4;i++) // Hat switches.
					{
						unsigned char mode;
						unsigned long bitdelta;
						
						// Load mode.  Override in order of complexity.
						if (DeviceExtension->Configurations[0][0][0])								
							mode = DeviceExtension->Configurations[0][0][0]->Hat[i].Mode;

						if (DeviceExtension->Configurations[switch1][0][0])								
							mode = DeviceExtension->Configurations[switch1][0][0]->Hat[i].Mode;
						if (DeviceExtension->Configurations[0][switch2][0])								
							mode = DeviceExtension->Configurations[0][switch2][0]->Hat[i].Mode;
						if (DeviceExtension->Configurations[0][0][shift])								
							mode = DeviceExtension->Configurations[0][0][shift]->Hat[i].Mode;
												
						if (DeviceExtension->Configurations[switch1][switch2][0])								
							mode = DeviceExtension->Configurations[switch1][switch2][0]->Hat[i].Mode;
						if (DeviceExtension->Configurations[switch1][0][shift])								
							mode = DeviceExtension->Configurations[switch1][0][shift]->Hat[i].Mode;
						if (DeviceExtension->Configurations[0][switch2][shift])								
							mode = DeviceExtension->Configurations[0][switch2][shift]->Hat[i].Mode;
												
						if (DeviceExtension->Configurations[switch1][switch2][shift])
							mode = DeviceExtension->Configurations[switch1][switch2][shift]->Hat[i].Mode;

						// Grab hat bitmap.
						pos=*(unsigned long*)&BufPtr[7];
						b = (pos >> ((i*4)+14))&0x0000000F;
						pos=*(unsigned long*)&Delta[7];
						bitdelta = (pos >> ((i*4)+14))&0x0000000F;
						
						if (mode&0x01) // 8-way
						{
							// reencode for ease of use.
							bitdelta = b ^ bitdelta;
							switch(b)
							{
								case 0x00:
									b = 0; break;
								case 0x01:
									b = 1; break;
								case 0x03:
									b = 2; break;
								case 0x02:
									b = 3; break;
								case 0x06:
									b = 4; break;
								case 0x04:
									b = 5; break;
								case 0x0C:
									b = 6; break;
								case 0x08:
									b = 7; break;
								case 0x09:
									b = 8; break;
							}
							switch(bitdelta)
							{
								case 0x00:
									bitdelta = 0; break;
								case 0x01:
									bitdelta = 1; break;
								case 0x03:
									bitdelta = 2; break;
								case 0x02:
									bitdelta = 3; break;
								case 0x06:
									bitdelta = 4; break;
								case 0x04:
									bitdelta = 5; break;
								case 0x0C:
									bitdelta = 6; break;
								case 0x08:
									bitdelta = 7; break;
								case 0x09:
									bitdelta = 8; break;
							}							
							
							
							if (bitdelta != b)
							{
								AddAllScripts(DeviceExtension,switch1,switch2,shift,&DeviceExtension->Configurations[0][0][0]->Hat[i].Pressed[b],LocalTime);
								AddAllScripts(DeviceExtension,switch1,switch2,shift,&DeviceExtension->Configurations[0][0][0]->Hat[i].Released[bitdelta],LocalTime);
							}
							else
							{
								// B held.
								;
							}
						}
						else // 4 way
						{
							for (pos=0;pos<4;pos++)
								if ((b>>pos) & 0x01) // Pressed
									if ((bitdelta>>pos) & 0x01) // Toggled
										AddAllScripts(DeviceExtension,switch1,switch2,shift,&DeviceExtension->Configurations[0][0][0]->Hat[i].Pressed[pos+1],LocalTime);
									else
									{
										// Held
										;
									}
								else
									if ((bitdelta>>pos) & 0x01) // Toggled
										AddAllScripts(DeviceExtension,switch1,switch2,shift,&DeviceExtension->Configurations[0][0][0]->Hat[i].Released[pos+1],LocalTime);
							
							if (b == 0)
								if (bitdelta)
									AddAllScripts(DeviceExtension,switch1,switch2,shift,&DeviceExtension->Configurations[0][0][0]->Hat[i].Pressed[0],LocalTime);
								else
								{
									// Position 0 held.
								}
							else
								if ((b ^ bitdelta)==0)
									AddAllScripts(DeviceExtension,switch1,switch2,shift,&DeviceExtension->Configurations[0][0][0]->Hat[i].Released[0],LocalTime);
						}
					}					

					// KdPrint(("Button Data: %02x %02x %02x %02x\n",BufPtr[7],BufPtr[8],BufPtr[9],BufPtr[10]));
					//BufPtr[9]=BufPtr[9] & 0xED;				

					RtlMoveMemory(&BufPtr[1],BufPtr,pBulkOrInterruptTransfer->TransferBufferLength);
			
					BufPtr[0]=0x01;
					
					pBulkOrInterruptTransfer->TransferBufferLength+=1;

					
					// Reencode Hat1 status.  Use shift as temp val.
					shift=0;
					pos=*(unsigned long*)&BufPtr[8];
//					DbgPrint("Hat reencoding:    pos = %X\n",pos);
					for(i=0;i<4;i++)
					{
						// Hat mode only valid for 000 config for now.
						if (DeviceExtension->Configurations[0][0][0]->Hat[i].Mode&0x02)
						{
							b = (pos >> ((i*4)+14))&0x0000000F;
//							DbgPrint("Hat reencoding:    Hat %d, b = %X\n",i,b);
							switch(b)
							{
								case 0x00:
									shift = 0; break;
								case 0x01:
									shift = 1; break;
								case 0x03:
									shift = 2; break;
								case 0x02:
									shift = 3; break;
								case 0x06:
									shift = 4; break;
								case 0x04:
									shift = 5; break;
								case 0x0C:
									shift = 6; break;
								case 0x08:
									shift = 7; break;
								case 0x09:
									shift = 8; break;
							}
						}
					}
//					DbgPrint("Hat reencoding:    shift = %d\n",shift);

					RtlCopyMemory(&BufPtr[8],&DeviceExtension->KeyboardStatus[21],4);
					BufPtr[0x0B]=(BufPtr[0x0B]&0x03)+(shift<<4);

					
					//DeviceExtension->KeyPressed=(BufPtr[8]&0x01);
            		
	        		//
					// Spawn timer if there isn't one already going.
       				//
       				Current = QueuePeek(&DeviceExtension->ActionQueue);
	       			if (( KeReadStateTimer( &DeviceExtension->Timer) ) && (Current))
	       			{
		           		KeSetTimer( &DeviceExtension->Timer, Current->Time, &DeviceExtension->TimerDPC );              
       				}
              		
					
					// Fix MDL.
//					if(Temp = (PUCHAR)MmGetSystemAddressForMdl(pBulkOrInterruptTransfer->TransferBufferMDL))
//						RtlCopyMemory(Temp,BufPtr,pBulkOrInterruptTransfer->TransferBufferLength);
				}
	
				break;

			default:
				// Do nothing.
				break;
		}
    
	}
	KdPrint(("JoyPendingActionCount: %d\n",DeviceExtension->pendingActionCount));
	return STATUS_SUCCESS;
}


VOID ExecuteScript( IN PKDPC Dpc, IN PVOID DeferredContext, IN PVOID SystemArgument1, IN PVOID SystemArgument2)
{
struct DEVICE_EXTENSION * DeviceExtension = (struct DEVICE_EXTENSION *)DeferredContext;
    LARGE_INTEGER SysTime;
    PQENTRY Current;

	// Check for new configuration.   Will not affect existing queue; moved here because of IRQL problems.
//	ReloadProfile(DeviceExtension);


	// Dequeue items.
	KeQuerySystemTime(&SysTime);
	SysTime = RtlLargeIntegerAdd(SysTime,RtlConvertLongToLargeInteger(20)); // Offset because of Windows' timer inaccuracy.
	
	Current = QueuePeek(&DeviceExtension->ActionQueue);
	while ((Current != NULL) && (RtlLargeIntegerGreaterThan(SysTime,Current->Time)))
	{
//		DbgPrint("Executing item:   Op: %d   Offset: %d   Time: %x %x\n",Current->Op,Current->Offset,Current->Time.HighPart,Current->Time.LowPart);		
		Current = QueuePop(&DeviceExtension->ActionQueue);
		if (Current->Op & 0x02)
		{
			Current->Offset+=(21*8)-1;
			DeviceExtension->NewData = DeviceExtension->NewData | 1; // New stick data
		}
		else
			DeviceExtension->NewData = DeviceExtension->NewData | 2; // New keybd data
		
		if (Current->Op & 0x01)
		{
			// Release.
			DeviceExtension->KeyboardStatus[(Current->Offset)/8]=DeviceExtension->KeyboardStatus[(Current->Offset)/8] & ~(1 << (Current->Offset%8));
		}
		else
		{
			// Press.
			DeviceExtension->KeyboardStatus[(Current->Offset)/8]=DeviceExtension->KeyboardStatus[(Current->Offset)/8] | (1 << (Current->Offset%8));
		}
		Current = QueuePeek(&DeviceExtension->ActionQueue);
	}
	
	
	// Reset timer.
	if (Current)
		KeSetTimer( &DeviceExtension->Timer, Current->Time, &DeviceExtension->TimerDPC );              
		

  	if (DeviceExtension->NewData)
  	{
	  		KeInsertQueueDpc(&DeviceExtension->ReturnDPC,NULL,NULL);
  	}

}



void CompleteRead( IN PKDPC Dpc, IN PVOID DeferredContext, IN PVOID SystemArgument1, IN PVOID SystemArgument2)
{
	struct DEVICE_EXTENSION * DeviceExtension = (struct DEVICE_EXTENSION *)DeferredContext;
	PIRP Irp = NULL;
    PIO_STACK_LOCATION IrpStack;
    ULONG bytesRead;
    UCHAR buffer[30];
    unsigned char switch1,switch2,shift;
    unsigned long b,i,pos;
    

	if (DeviceExtension)
	{
		KIRQL OldIrqL;
				
		KeAcquireSpinLockAtDpcLevel(&DeviceExtension->ReadLock);
		Irp = DeviceExtension->ReadIrp;
		DeviceExtension->ReadIrp = NULL;
		KeReleaseSpinLockFromDpcLevel(&DeviceExtension->ReadLock);
	}

	if (Irp)
	{
		IrpStack = IoGetCurrentIrpStackLocation(Irp);
		
		if (DeviceExtension->NewData & 0x01)
		{	
			KdPrint(("Completing joystick read..\n"));
			
			buffer[0]=0x01;
			RtlCopyMemory(&buffer[1],DeviceExtension->LastPosition,0x0b);
			RtlCopyMemory(&buffer[8],&DeviceExtension->KeyboardStatus[21],4);
			
			// Reencode Hat1 status.  Use shift as temp val.
			shift=0;
			pos=*(unsigned long*)&buffer[8];
//			DbgPrint("Hat reencoding:    pos = %X\n",pos);
			for(i=0;i<4;i++)
			{
				// Hat mode only valid for 000 config for now.
				if (DeviceExtension->Configurations[0][0][0]->Hat[i].Mode == 0x03)
				{
					b = (pos >> ((i*4)+14))&0x0000000F;
//					DbgPrint("Hat reencoding:    Hat %d, b = %X\n",i,b);
					switch(b)
					{
						case 0x00:
							shift = 0; break;
						case 0x01:
							shift = 1; break;
						case 0x03:
							shift = 2; break;
						case 0x02:
							shift = 3; break;
						case 0x06:
							shift = 4; break;
						case 0x04:
							shift = 5; break;
						case 0x0C:
							shift = 6; break;
						case 0x08:
							shift = 7; break;
						case 0x09:
							shift = 8; break;
					}
				}
			}
//			DbgPrint("Hat reencoding:    shift = %d\n",shift);
	
	
			buffer[0x0B]=(buffer[0x0B]&0x03)+(shift<<4);
			bytesRead=0x0c;
			DeviceExtension->NewData=DeviceExtension->NewData&0xFE;
		}
		else	
		{
			
		    KdPrint(("Completing keyboard read..\n"));
	 			
			RtlCopyMemory(&buffer[1],DeviceExtension->KeyboardStatus,29);
			buffer[0]=0x02;
		
			RtlFillMemory(&buffer[20],4,0x00);	
	
			// Complete our IO report.
	    	bytesRead = 30;			
	    	DeviceExtension->NewData=DeviceExtension->NewData&0xFD;
		}
		    	
		if (bytesRead > IrpStack->Parameters.DeviceIoControl.OutputBufferLength)
	    {
	       	bytesRead = IrpStack->Parameters.DeviceIoControl.OutputBufferLength;
	    }
	
		RtlCopyMemory((PUCHAR) Irp->UserBuffer, buffer, bytesRead);
	    Irp->IoStatus.Information = bytesRead;
	    Irp->IoStatus.Status = STATUS_SUCCESS;
		
	    IoCompleteRequest(Irp, IO_NO_INCREMENT);
	    DecrementPendingActionCount(DeviceExtension);
	}
}


