// QUEUE.C
//
// Handles the timer queue for the X36 scripting engine.
// 
// includes:
//
// queueinit: inits queue.
// scriptadd: adds a whole script to the queue.
// pop: removes the first item from the queue.
// peek: returns the first item in the queue without destroying it
// queueclear: destroys a queue.


#include <WDM.H>
#include "filter.h"


NTSTATUS QueueInit(PXQUEUE MyQueue)
{
	unsigned int i;
	
	if (MyQueue == NULL)
		return STATUS_INVALID_PARAMETER;
	
	MyQueue->Head = NULL;
	KeInitializeSpinLock(&MyQueue->Lock);
	MyQueue->Next = 0; MyQueue->Last = 0;
	for (i=0; i<256; i++)
		MyQueue->Free[i]=(unsigned char)i;
	// MyQueue->Buffer = ExAllocatePool(PagedPool, sizeof(QENTRY[255]));
	
	return STATUS_SUCCESS;
}

NTSTATUS AddScript(PXQUEUE MyQueue, PSCRIPT MyScript, LARGE_INTEGER RefTime)
{
	PSCRIPT SCurrent;
	PQENTRY QCurrent, QPrev, QNew;
	KIRQL OldIrqL;
	
//	DbgPrint("Entering AddScript.\n");
	if (MyQueue == NULL)
		return STATUS_INVALID_PARAMETER;
		
	KeAcquireSpinLockAtDpcLevel(&MyQueue->Lock);
//	DbgPrint("Spinlock aquired.\n");

	QPrev = NULL;
	QCurrent = MyQueue->Head;
	SCurrent = MyScript;
	
//	DbgPrint("Beginning loop.\n");
	while (SCurrent != NULL)
	{
		if (MyQueue->Next - MyQueue->Last < 255)
		{
//			DbgPrint("Assigning new space in buffer..");
			QNew = &MyQueue->Buffer[MyQueue->Free[MyQueue->Next]];
			MyQueue->Next++;
//			DbgPrint("..assigned.\n");
			RtlZeroMemory(QNew,sizeof(QENTRY));
//			DbgPrint("New entry cleared.\n");
			QNew->Op = SCurrent->Op;
			QNew->Offset = SCurrent->Offset;
			QNew->Time = RtlLargeIntegerAdd(RtlConvertLongToLargeInteger(SCurrent->Delay),RefTime);
//			DbgPrint("Queuing item:   Op: %d   Offset: %d   Time: %x %x\n",QNew->Op,QNew->Offset,QNew->Time.HighPart,QNew->Time.LowPart);
			
			if (QCurrent == NULL)
				MyQueue->Head = QCurrent = QNew;
			else
			{
				while((QCurrent != NULL) && (RtlLargeIntegerGreaterThan(QNew->Time,QCurrent->Time)))
				{
					QPrev = QCurrent;
					QCurrent = QCurrent -> Next;
				}
				
				
				if (QPrev != NULL)
				{
					QNew->Next = QCurrent;
					QPrev->Next = QNew;
				}
				else
				{
					MyQueue->Head = QNew;
					QNew->Next = QCurrent;				
				}
				QCurrent=QNew;
			}
			SCurrent= SCurrent->Next;			
		}
		else
		{
			DbgPrint("Addscript aborted due to insufficent resources.\n");
			KeReleaseSpinLockFromDpcLevel(&MyQueue->Lock);
//			DbgPrint("Spinlock released.\n");
			return STATUS_INSUFFICIENT_RESOURCES;
		}
	}
//	DbgPrint("Exiting loop.\n");			
	KeReleaseSpinLockFromDpcLevel(&MyQueue->Lock);
//	DbgPrint("Spinlock released.\n");
	return STATUS_SUCCESS;
}



PQENTRY QueuePop(PXQUEUE MyQueue)
{
	PQENTRY QCurrent, QReturn;
	KIRQL OldIrqL;
	
	if (MyQueue == NULL)
		return NULL;
		
	KeAcquireSpinLockAtDpcLevel(&MyQueue->Lock);
	
	QCurrent = MyQueue->Head;
	if (QCurrent != NULL)
	{
		MyQueue->Head = QCurrent->Next;
		MyQueue->Free[MyQueue->Last]=(QCurrent - &MyQueue->Buffer[0])/sizeof(QENTRY);
		MyQueue->Last++;
	}
	
	KeReleaseSpinLockFromDpcLevel(&MyQueue->Lock);
	return QCurrent;
}

PQENTRY QueuePeek(PXQUEUE MyQueue)
{
	PQENTRY QCurrent;
	KIRQL OldIrqL;
	
	if (MyQueue == NULL)
		return NULL;
		
	KeAcquireSpinLockAtDpcLevel(&MyQueue->Lock);
	
	QCurrent = MyQueue->Head;
	
	KeReleaseSpinLockFromDpcLevel(&MyQueue->Lock);
	return QCurrent;
}

void QueueClear(PXQUEUE MyQueue)
{
	PQENTRY QCurrent;
	KIRQL OldIrqL;	
	
	if (MyQueue == NULL)
		return;
		
	KeAcquireSpinLockAtDpcLevel(&MyQueue->Lock);

	while(MyQueue->Head != NULL)
	{
		QCurrent = MyQueue->Head;
		MyQueue->Head = QCurrent->Next;
		MyQueue->Free[MyQueue->Last]=(QCurrent - &MyQueue->Buffer[0])/sizeof(QENTRY);		
		MyQueue->Last++;
	}
	
	KeReleaseSpinLockFromDpcLevel(&MyQueue->Lock);
}


void AddAllScripts(struct DEVICE_EXTENSION *devExt, unsigned char mode, unsigned char aux, unsigned char shift, void* Config0Pos, LARGE_INTEGER Time)
{
	unsigned int Delta;
	unsigned char i,j,k;
	
	Delta = (unsigned char*)Config0Pos - (unsigned char*)devExt->Configurations[0][0][0];
//	DbgPrint("AddAllScripts delta: %d   mode: %d   aux: %d   shift: %d\n",Delta,mode,aux,shift);
//	DbgPrint("AAS Computed: %X   Actual: %X\n",(((unsigned char*)devExt->Configurations[0][0][0])+Delta),&(devExt->Configurations[0][0][0]->Button[0].Released));

	for(i=0;i<2;i++)
		for(j=0;j<2;j++)
			for(k=0;k<2;k++)
				if (devExt->Configurations[i*mode][aux*j][shift*k])
				{
//					DbgPrint("AddAllScripts adding:  delta: %d   mode: %d   aux: %d   shift: %d\n",Delta,mode*i,aux*j,shift*k);
					AddScript(&devExt->ActionQueue,*(PSCRIPT*)(((unsigned char*)devExt->Configurations[i*mode][aux*j][shift*k])+Delta), Time);
				}
				
//	DbgPrint("AddAllScripts ending.\n");
}